/* $Id: shbufmodule.c,v 1.2 2003/09/11 16:42:12 poettering Exp $ */

/*
 * This file is part of python-shbuf. 
 *
 * python-shbuf is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * python-shbuf is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with python-shbuf; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 */

#include <Python.h>
#include <libshbuf/shbuf.h>
#include <libshbuf/shbuferr.h>

static PyObject *shbufError;

staticforward PyTypeObject shbufType;
staticforward PyMethodDef shbufObjectMethods[];

typedef struct {
    PyObject_HEAD
    shbuf* sb;
} shbufObject;

static PyObject *pyshbuf_open(PyObject *self, PyObject *args) {
    key_t key;
    shbufObject* sbo;
    shbuf *sb;
    
    if (!PyArg_ParseTuple(args, "i", &key))
        return NULL;

    if (!(sb = shbuf_open(key))) {
        PyErr_SetString(shbufError, shbuf_strerror2());
        return NULL;
    }

    if (shbuf_notify_enable(sb, 1) != 0) {
        shbuf_free(sb);
        PyErr_SetString(shbufError, shbuf_strerror2());
        return NULL;
    }
    
    sbo = PyObject_New(shbufObject, &shbufType);
    sbo->sb = sb;
    
    return (PyObject*) sbo;
}

static void pyshbuf_dealloc(PyObject *self) {
    shbufObject *sbo = (shbufObject*) self;

    shbuf_notify_enable(sbo->sb, 0);
    shbuf_free(sbo->sb);
        
    PyObject_Del(self);
}

static PyObject *pyshbuf_repr(PyObject *self) {
    shbufObject *sbo = (shbufObject*) self;
    char buf[100];

    sprintf(buf, "<shbuf object [key=0x%x, size=%ld]>",
            shbuf_get_key(sbo->sb),
            shbuf_get_size(sbo->sb));
    
    return PyString_FromString(buf);
}

static PyObject *pyshbuf_getattr(PyObject *self, char *name) {
    return Py_FindMethod(shbufObjectMethods, self, name);
}

#ifndef MIN
#define MIN(x,y) ((x)<(y)?(x):(y))
#endif

static PyObject *pyshbuf_write(PyObject *self, PyObject *args) {
    shbufObject *sbo = (shbufObject*) self;
    unsigned char *buf;
    unsigned long l;
    
    unsigned char *wbuf;
    unsigned long wl;

    if (!PyArg_ParseTuple(args, "s#", &wbuf, &wl))
        return NULL;

    while (!(buf = shbuf_get_write_pointer(sbo->sb, &l)))
        if (shbuf_wait(sbo->sb) != 0) {
            PyErr_SetString(shbufError, shbuf_strerror2());
            return NULL;
        }

    l = MIN(wl,l);

    memcpy(buf, wbuf, l);
    shbuf_inc_write_pointer(sbo->sb, l);
    
    if (shbuf_notify(sbo->sb) != 0) {
        PyErr_SetString(shbufError, shbuf_strerror2());
        return NULL;
    }
    
    return Py_BuildValue("i", l);
}

static PyObject *pyshbuf_is_full(PyObject *self, PyObject *args) {
    shbufObject *sbo = (shbufObject*) self;
    int r;

    if (!PyArg_ParseTuple(args, ""))
        return NULL;

    r = shbuf_is_full(sbo->sb);

    if (r < 0) {
        PyErr_SetString(shbufError, shbuf_strerror2());
        return NULL;
    }
    
    return Py_BuildValue("i", r);
}

static PyObject *pyshbuf_is_empty(PyObject *self, PyObject *args) {
    shbufObject *sbo = (shbufObject*) self;
    int r;

    if (!PyArg_ParseTuple(args, ""))
        return NULL;

    r = shbuf_is_empty(sbo->sb);

    if (r < 0) {
        PyErr_SetString(shbufError, shbuf_strerror2());
        return NULL;
    }
    
    return Py_BuildValue("i", r);
}

static PyObject *pyshbuf_reset(PyObject *self, PyObject *args) {
    shbufObject *sbo = (shbufObject*) self;

    if (!PyArg_ParseTuple(args, ""))
        return NULL;

    if (shbuf_reset(sbo->sb) != 0) {
        PyErr_SetString(shbufError, shbuf_strerror2());
        return NULL;
    }
    
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *pyshbuf_zero(PyObject *self, PyObject *args) {
    shbufObject *sbo = (shbufObject*) self;

    if (!PyArg_ParseTuple(args, ""))
        return NULL;

    if (shbuf_zero(sbo->sb) != 0) {
        PyErr_SetString(shbufError, shbuf_strerror2());
        return NULL;
    }
    
    Py_INCREF(Py_None);
    return Py_None;  
}


static PyObject *pyshbuf_get_select_fd(PyObject *self, PyObject *args) {
    int fd;
    shbufObject *sbo = (shbufObject*) self;
    
    if (!PyArg_ParseTuple(args, ""))
        return NULL;

    if ((fd = shbuf_get_select_fd(sbo->sb)) == -1) {
        PyErr_SetString(shbufError, shbuf_strerror2());
        return NULL;
    }
    
    return Py_BuildValue("i", fd);
}

static PyObject *pyshbuf_post_select(PyObject *self, PyObject *args) {
    shbufObject *sbo = (shbufObject*) self;
    
    if (!PyArg_ParseTuple(args, ""))
        return NULL;

    if (shbuf_post_select(sbo->sb) != 0) {
        PyErr_SetString(shbufError, shbuf_strerror2());
        return NULL;
    }
    
    Py_INCREF(Py_None);
    return Py_None;  
}

static PyObject *pyshbuf_notify(PyObject *self, PyObject *args) {
    shbufObject *sbo = (shbufObject*) self;
    
    if (!PyArg_ParseTuple(args, ""))
        return NULL;

    if (shbuf_notify(sbo->sb) != 0) {
        PyErr_SetString(shbufError, shbuf_strerror2());
        return NULL;
    }
    
    Py_INCREF(Py_None);
    return Py_None;  
}

static PyTypeObject shbufType = {
    PyObject_HEAD_INIT(NULL)
    0,
    "shbuf",
    sizeof(shbufObject),
    0,
    pyshbuf_dealloc, /*tp_dealloc*/
    0,          /*tp_print*/
    pyshbuf_getattr,          /*tp_getattr*/
    0,          /*tp_setattr*/
    0,          /*tp_compare*/
    pyshbuf_repr,          /*tp_repr*/
    0,          /*tp_as_number*/
    0,          /*tp_as_sequence*/
    0,          /*tp_as_mapping*/
    0,          /*tp_hash */
};

static PyMethodDef shbufObjectMethods[] = {
    {"write",  (PyCFunction) pyshbuf_write, METH_VARARGS, "Write a string to the shared buffer." },
    {"is_full",  (PyCFunction) pyshbuf_is_full, METH_VARARGS, "Check whether it is possible to write data to the shared buffer without blocking." },
    {"is_empty",  (PyCFunction) pyshbuf_is_empty, METH_VARARGS, "Check whether it is possible to write data to the shared buffer without blocking." },
    {"reset",    (PyCFunction) pyshbuf_reset, METH_VARARGS, "Empty the shared buffer" },
    {"zero",  (PyCFunction) pyshbuf_zero, METH_VARARGS, "Empty the shared buffer and zero it out" },
    {"get_select_fd", (PyCFunction) pyshbuf_get_select_fd, METH_VARARGS, "Return a unix file descriptor, a program may select() on" },
    {"post_select",  (PyCFunction) pyshbuf_post_select, METH_VARARGS, "Clean up the shbuf object after having used select()" },
    {"notify",  (PyCFunction) pyshbuf_notify, METH_VARARGS, "Notify the other side of the buffer about changes." },
    {NULL, NULL}
};

static PyMethodDef shbufModuleMethods[] = {
    {"open",  (PyCFunction) pyshbuf_open, METH_VARARGS },
    {NULL, NULL}
};

DL_EXPORT(void) initshbuf(void) 
{
    PyObject *m, *d;
    shbufType.ob_type = &PyType_Type;
    m = Py_InitModule("shbuf", shbufModuleMethods);
    d = PyModule_GetDict(m);
    shbufError = PyErr_NewException("shbuf.error", NULL, NULL);
    PyDict_SetItemString(d, "error", shbufError);
    PyDict_SetItemString(d, "__doc__", PyString_FromString("libshbuf encapuslation for Python"));
}

